@@ -10,7 +10,7 @@
 #ifndef LLD_READER_WRITER_YAML_CONTEXT_H
 #define LLD_READER_WRITER_YAML_CONTEXT_H
 
-#include "lld/Core/LLVM.h"
+#include "lld/Common/LLVM.h"
 #include <functional>
 #include <memory>
 #include <vector>
@@ -18,6 +18,7 @@
 namespace lld {
 class File;
 class LinkingContext;
+class Registry;
 namespace mach_o {
 namespace normalized {
 struct NormalizedFile;
