@@ -282,7 +282,7 @@ public:
 
 private:
   llvm::Error perform(SimpleFile &mergedFile) override {
-    DEBUG(llvm::dbgs() << "MachO Compact Unwind pass\n");
+    LLVM_DEBUG(llvm::dbgs() << "MachO Compact Unwind pass\n");
 
     std::map<const Atom *, CompactUnwindEntry> unwindLocs;
     std::map<const Atom *, const Atom *> dwarfFrames;
@@ -319,7 +319,7 @@ private:
 
     // Finally, we can start creating pages based on these entries.
 
-    DEBUG(llvm::dbgs() << "  Splitting entries into pages\n");
+    LLVM_DEBUG(llvm::dbgs() << "  Splitting entries into pages\n");
     // FIXME: we split the entries into pages naively: lots of 4k pages followed
     // by a small one. ld64 tried to minimize space and align them to real 4k
     // boundaries. That might be worth doing, or perhaps we could perform some
@@ -336,11 +336,13 @@ private:
       pages.back().entries = remainingInfos.slice(0, entriesInPage);
       remainingInfos = remainingInfos.slice(entriesInPage);
 
-      DEBUG(llvm::dbgs()
-            << "    Page from " << pages.back().entries[0].rangeStart->name()
-            << " to " << pages.back().entries.back().rangeStart->name() << " + "
-            << llvm::format("0x%x", pages.back().entries.back().rangeLength)
-            << " has " << entriesInPage << " entries\n");
+      LLVM_DEBUG(llvm::dbgs()
+                 << "    Page from "
+                 << pages.back().entries[0].rangeStart->name() << " to "
+                 << pages.back().entries.back().rangeStart->name() << " + "
+                 << llvm::format("0x%x",
+                                 pages.back().entries.back().rangeLength)
+                 << " has " << entriesInPage << " entries\n");
     } while (!remainingInfos.empty());
 
     auto *unwind = new (_file.allocator())
@@ -360,7 +362,7 @@ private:
       const SimpleFile &mergedFile,
       std::map<const Atom *, CompactUnwindEntry> &unwindLocs,
       std::vector<const Atom *> &personalities, uint32_t &numLSDAs) {
-    DEBUG(llvm::dbgs() << "  Collecting __compact_unwind entries\n");
+    LLVM_DEBUG(llvm::dbgs() << "  Collecting __compact_unwind entries\n");
 
     for (const DefinedAtom *atom : mergedFile.defined()) {
       if (atom->contentType() != DefinedAtom::typeCompactUnwindInfo)
@@ -369,14 +371,15 @@ private:
       auto unwindEntry = extractCompactUnwindEntry(atom);
       unwindLocs.insert(std::make_pair(unwindEntry.rangeStart, unwindEntry));
 
-      DEBUG(llvm::dbgs() << "    Entry for " << unwindEntry.rangeStart->name()
-                         << ", encoding="
-                         << llvm::format("0x%08x", unwindEntry.encoding));
+      LLVM_DEBUG(llvm::dbgs() << "    Entry for "
+                              << unwindEntry.rangeStart->name() << ", encoding="
+                              << llvm::format("0x%08x", unwindEntry.encoding));
       if (unwindEntry.personalityFunction)
-        DEBUG(llvm::dbgs() << ", personality="
-                           << unwindEntry.personalityFunction->name()
-                           << ", lsdaLoc=" << unwindEntry.lsdaLocation->name());
-      DEBUG(llvm::dbgs() << '\n');
+        LLVM_DEBUG(llvm::dbgs()
+                   << ", personality="
+                   << unwindEntry.personalityFunction->name()
+                   << ", lsdaLoc=" << unwindEntry.lsdaLocation->name());
+      LLVM_DEBUG(llvm::dbgs() << '\n');
 
       // Count number of LSDAs we see, since we need to know how big the index
       // will be while laying out the section.
@@ -454,7 +457,7 @@ private:
       const std::map<const Atom *, const Atom *> &dwarfFrames) {
     std::vector<CompactUnwindEntry> unwindInfos;
 
-    DEBUG(llvm::dbgs() << "  Creating __unwind_info entries\n");
+    LLVM_DEBUG(llvm::dbgs() << "  Creating __unwind_info entries\n");
     // The final order in the __unwind_info section must be derived from the
     // order of typeCode atoms, since that's how they'll be put into the object
     // file eventually (yuck!).
@@ -465,10 +468,10 @@ private:
       unwindInfos.push_back(finalizeUnwindInfoEntryForAtom(
           atom, unwindLocs, personalities, dwarfFrames));
 
-      DEBUG(llvm::dbgs() << "    Entry for " << atom->name()
-                         << ", final encoding="
-                         << llvm::format("0x%08x", unwindInfos.back().encoding)
-                         << '\n');
+      LLVM_DEBUG(llvm::dbgs()
+                 << "    Entry for " << atom->name() << ", final encoding="
+                 << llvm::format("0x%08x", unwindInfos.back().encoding)
+                 << '\n');
     }
 
     return unwindInfos;
