@@ -13,33 +13,50 @@
 #include "llvm/ADT/StringRef.h"
 #include "llvm/ADT/StringSet.h"
 #include "llvm/BinaryFormat/Wasm.h"
-
-#include "Symbols.h"
-
-using llvm::wasm::WasmGlobal;
+#include "llvm/Support/CachePruning.h"
 
 namespace lld {
 namespace wasm {
 
 struct Configuration {
   bool AllowUndefined;
-  bool CheckSignatures;
+  bool CompressRelocations;
   bool Demangle;
-  bool EmitRelocs;
+  bool DisableVerify;
+  bool ExportAll;
+  bool ExportDynamic;
+  bool ExportTable;
+  bool GcSections;
   bool ImportMemory;
+  bool SharedMemory;
+  bool ImportTable;
+  bool MergeDataSegments;
+  bool Pie;
+  bool PrintGcSections;
   bool Relocatable;
+  bool SaveTemps;
+  bool Shared;
   bool StripAll;
   bool StripDebug;
+  bool StackFirst;
   uint32_t GlobalBase;
   uint32_t InitialMemory;
   uint32_t MaxMemory;
   uint32_t ZStackSize;
+  unsigned LTOPartitions;
+  unsigned LTOO;
+  unsigned Optimize;
+  unsigned ThinLTOJobs;
   llvm::StringRef Entry;
   llvm::StringRef OutputFile;
+  llvm::StringRef ThinLTOCacheDir;
 
   llvm::StringSet<> AllowUndefinedSymbols;
   std::vector<llvm::StringRef> SearchPaths;
-  Symbol *StackPointerSymbol = nullptr;
+  llvm::CachePruningPolicy ThinLTOCachePolicy;
+
+  // True if we are creating position-independent code.
+  bool Pic;
 };
 
 // The only instance of Configuration struct.
